%% @author jstypka <jasieek@student.agh.edu.pl>
%% @version 1.1
%% @doc This is the main module of sequential model. It handles starting the system and cleaning after work

-module(mas_sequential).
-export([start/3]).

-include ("./erlang-mas-master/include/mas.hrl").

-type agent() :: agent().
-type sim_params() :: mas:sim_params().

%% ====================================================================
%% API functions
%% ====================================================================

-spec start(Time::pos_integer(), sim_params(), config()) -> [agent()].
start(Time, SP, Cf = #config{islands = Islands}) ->
    mas_misc_util:seed_random(),
    mas_misc_util:clear_inbox(),
    mas_misc_util:initialize_subscriptions(lists:seq(1, Islands), Cf),
    mas_topology:start_link(self(), Islands, Cf#config.topology),
    InitIslands = [mas_misc_util:generate_population(SP, Cf)
                   || _ <- lists:seq(1, Islands)],
    {_Time, Result} = timer:tc(fun loop/4,
                               [InitIslands,
                                Time,
                                SP,
                                Cf]),
    mas_misc_util:close_subscriptions(lists:seq(1, Islands), Cf),
    mas_topology:close(),
    lists:flatten(Result).

%% ====================================================================
%% Internal functions
%% ====================================================================

%% @doc The main island process loop.
%% A new generation of the population is created in every iteration.
-spec loop([island()], pos_integer(), sim_params(), config()) -> [agent()].
loop(Islands, Time, SP, Cf) ->
    Tag = fun(Island) ->
                  [{mas_misc_util:behaviour_proxy(Agent,
                                                  SP,
                                                  Cf), Agent}
                   || Agent <- Island]
          end,

    Groups = [mas_misc_util:group_by(Tag(I))
              || I <- Islands],

    Migrants = [seq_migrate(lists:keyfind(migration, 1, Island), Nr)
                 || {Island, Nr} <- lists:zip(Groups,
                                              lists:seq(1, length(Groups)))],

    NewGroups = 
	[[mas_misc_util:meeting_proxy(Activity, mas_sequential, SP, Cf) || Activity <- I]
	    || I <- Groups],

    WithMigrants = append(lists:flatten(Migrants), NewGroups),

    NewIslands = [mas_misc_util:shuffle(lists:flatten(I))
                  || I <- WithMigrants],

    case os:timestamp() < Time of
        true ->
            loop(NewIslands, Time, SP, Cf);
        false ->
            NewIslands
    end.

-spec seq_migrate(false | {migration, [agent()]}, pos_integer()) ->
                         [{migration, [agent()]}].
seq_migrate(false, _) ->
    [];

seq_migrate({migration, Agents}, From) ->
    Destinations = [{mas_topology:getDestination(From), Agent}
                    || Agent <- Agents],
    mas_misc_util:group_by(Destinations).


-spec append([{pos_integer(), [agent()]}], [list(agent())]) -> [list(agent())].
append([], Islands) ->
    Islands;

append([{Destination, Immigrants} | T], Islands) ->
    NewIslands = mas_misc_util:map_index(Immigrants,
                                         Destination,
                                         Islands,
                                         fun lists:append/2),
    append(T, NewIslands).

tgm(TagFun, MigrantsFun, GroupFun) ->
    fun(Agents) ->
            Tagged = lists:map(TagFun, Agents),
            Migrants = lists:map(MigrantsFun, Tagged),
            _Groups = GroupFun(Migrants)
    end.

%% @doc Evaluates a given solution by computing the Rastrigin function.
evaluation(B, _SP) ->
    - lists:foldl(fun(X, Sum) -> Sum + 10 + X*X - 10*math:cos(2*math:pi()*X) end, 0.0, B).
